class CryptException(Exception):
	"""Base class for all exceptions in the module."""
	pass


class FileParserError(CryptException):
	"""Used when there is an error parsing and encrypted file."""
	pass



import base64
import binascii
import json

import typing

from .types import RSAKeySet
from .low_level import encrypt, decrypt


# TODO: Think about adding signing to these helpers, or switching up encryption modes to something with verification.


def encrypt_string(message: str, rsa_key: RSAKeySet) -> str:
	"""Encrypt the given string."""
	# Encode it down to ASCII
	message = message.encode()

	# Encrypt the message
	encrypted_message = encrypt(message, rsa_key)

	# Wrap binary in base64
	encrypted_message = base64.b64encode(encrypted_message)

	# Convert the bytes back into a python str
	encrypted_message = encrypted_message.decode('utf-8')

	return encrypted_message


def decrypt_string(encrypted_message: str, private_keys: RSAKeySet) -> str:
	"""Decrypt the given string."""
	# Encode it down to ASCII
	encrypted_message = encrypted_message.encode()

	# Strip base64, back to binary
	encrypted_message = base64.b64decode(encrypted_message)

	# Decrypt the message
	message = decrypt(encrypted_message, private_keys).decode()

	return message


def encrypt_object(message: typing.Any, rsa_key: RSAKeySet) -> str:
	"""Encrypt an python object. Must be json serializable."""
	# Wrap it in JSON
	message = json.dumps(message)

	# Encode it down to ASCII
	message = message.encode()

	# Encrypt the message
	encrypted_message = encrypt(message, rsa_key)

	# Wrap binary in base64
	encrypted_message = base64.b64encode(encrypted_message)

	# Convert the bytes back into a python str
	encrypted_message = encrypted_message.decode('utf-8')

	return encrypted_message


def decrypt_object(encrypted_message: str, private_keys: RSAKeySet, ignore_failure: bool = False) -> typing.Any:
	"""Decrypt the message back into python objects."""
	# Encode it down to ASCII
	encrypted_message = encrypted_message.encode()

	# Strip base64, back to binary
	try:
		encrypted_message = base64.b64decode(encrypted_message)
	except binascii.Error:
		if ignore_failure:
			return encrypted_message.decode('utf-8')
		raise

	# Decrypt the message
	message = decrypt(encrypted_message, private_keys).decode()

	# Unwrap the JSON
	message = json.loads(message)

	return message



import sys

try:
	from cfcrypt.exceptions import CryptException
except ImportError:
	CryptException = RuntimeError


ENABLE_STACK_TRACE = False


class RSAException(CryptException):
	pass


class CryptExceptionNoStack(RSAException):
	"""The exception handler is hooked to not show the stack trace for these exceptions."""
	pass


class EncryptionFailed(CryptExceptionNoStack):
	pass


class DecryptionFailed(CryptExceptionNoStack):
	pass


class InvalidSignature(CryptExceptionNoStack):
	pass


current_handler = sys.excepthook


def exception_handler(exception_type, exception, traceback):
	if not ENABLE_STACK_TRACE and issubclass(exception_type, CryptExceptionNoStack):
		print("%s: %s" % (exception_type.__name__, exception), file=sys.stderr)
		sys.exit(1)
	return current_handler(exception_type, exception, traceback)


# noinspection SpellCheckingInspection
sys.excepthook = exception_handler



import os
import stat
from getpass import getpass
from pathlib import Path
from typing import Optional, TypeVar
import string
import secrets

# import cfjson

from cfcrypt import CryptFileBinaryIO, CryptFileTextIO
from cfcrypt.rsa.types import RSAPrivateKeySet, RSAPrivateKey
from cfcrypt.rsa.key_helpers import generate_private_key, generate_private_pem, pem_file_to_private_key, \
	private_key_to_pem, private_key_to_pem_file


PathLike = TypeVar("PathLike", str, bytes, os.PathLike)


class MetaKeyRing(type):
	"""Singleton for accessing the key ring."""

	_instances = {}

	def __call__(cls, *args, **kwargs):
		if cls not in cls._instances:
			cls._instances[cls] = super(MetaKeyRing, cls).__call__(*args, **kwargs)
		return cls._instances[cls]

	def __init__(cls, *args, **kwargs):
		super(MetaKeyRing, cls).__init__(*args, **kwargs)


def confirm_password(msg: str, confirmation: bool = True) -> str:
	"""Get a password from the user, with confirmation.

	:type msg: Message to print before the prompts.
	:param confirmation: If True the user will be asked to confirm the password.

	:meta private:
	"""
	while True:
		print(msg)
		password = getpass('Password: ')
		if not confirmation:
			break
		password2 = getpass('Confirm Password: ')
		if password == password2:
			break
		print('Passwords were different. Try again.')
	return password


class KeyRingBase(object, metaclass=MetaKeyRing):
	"""KeyRing helps you keep track of your applications RSA keys in a secure way.

	:meta private:
	"""

	@classmethod
	def create_key_ring_interactive(cls):
		"""Create a new KeyRing.

		Note: Must be called from an interactive prompt to allow for password entry.
		"""
		if 'CRYPT_KEYRING' not in os.environ:
			raise RuntimeError('KeyRing path not set in environment variable CRYPT_KEYRING')

		kek_path = KeyRingBase.main_key_file()
		if kek_path.exists():
			raise RuntimeError('Cannot create a new key ring as one already exists here: {}'.format(kek_path))

		print('Creating new KeyRing')
		password = confirm_password(
			'Enter a new password.\n'
			'This will be used to unlock the KeyRingBase. Choose a strong password with a length of at least 16.')

		print('Creating new master key.')
		pem_data = generate_private_pem(4096, password)
		with kek_path.open('wb') as fh:
			fh.write(pem_data)
		os.chmod(kek_path, stat.S_IREAD)

		instance = cls(password=password)
		instance._dump()
		return instance

	@classmethod
	def create_key_ring(cls):
		"""Create a new KeyRing.

		Note: During the creation an access token will be displayed. This is the only time it can been seen.
		Make sure you get a copy of it to access the KeyRing in future.
		"""

		if 'CRYPT_KEYRING' not in os.environ:
			raise RuntimeError('KeyRing path not set in environment variable CRYPT_KEYRING')

		kek_path = KeyRingBase.main_key_file()
		if kek_path.exists():
			raise RuntimeError('Cannot create a new key ring as one already exists here: {}'.format(kek_path))

		print('Creating new KeyRing')
		password = password_generator(50)

		print('Creating new master key.')
		private_key = generate_private_key(4096)
		private_key_to_pem_file(kek_path, private_key, password)
		os.chmod(kek_path, stat.S_IREAD)

		print(
			f'KeyRing has been created in {KeyRingBase.directory()}\n'
			f'Your access token is:\n'
			f'{password}\n'
			f'The token goes into the environmental variable CRYPT_KEYRING_TOKEN\n'
			f'Keep a backup of the token in a safe place as KeyRing does not keep a copy of it.')

		instance = cls(password=password)
		instance._dump()
		return instance

	@classmethod
	def change_password_interactive(cls):
		"""Change the password or access token of the KeyRing.

		Note: Must be called from an interactive prompt to allow for password entry.
		"""
		kek_path = KeyRingBase.main_key_file()
		if not kek_path.exists():
			raise RuntimeError('Could not find the key: {}'.format(kek_path))

		current_password = getpass('Current Password: ')
		kek = pem_file_to_private_key(kek_path, current_password.encode())

		password = confirm_password(
			'Enter a new password.\n'
			'This will be used to unlock the KeyRingBase. Choose a strong password with a length of at least 16.')

		temp = kek_path.with_suffix('.temp')
		try:
			with temp.open('wb') as fh:
				fh.write(private_key_to_pem(kek, password))
		except Exception:
			temp.unlink(missing_ok=True)
			raise
		os.chmod(kek_path, stat.S_IREAD | stat.S_IWRITE)
		temp.replace(kek_path)
		os.chmod(kek_path, stat.S_IREAD)

		return cls(password)

	@staticmethod
	def directory() -> Path:
		"""Get the working directory of the KeyRing.

		:meta private:
		"""
		working_dir = Path(os.environ.get('CRYPT_KEYRING', '..'))
		working_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
		return working_dir

	@staticmethod
	def main_key_file() -> Path:
		"""The location of the KeyRing's main RSAPrivateKey. All data is secured with this key.

		:meta private:
		"""
		return KeyRingBase.directory() / 'main.pem'

	@property
	def main_key(self) -> RSAPrivateKey:
		"""The main encryption key used to encrypt all other keys.

		:meta private:
		"""
		if self._kek is None:
			self._kek = pem_file_to_private_key(self.main_key_file(), self.password)
		return self._kek

	def __init__(self, namespace=None, password=None):
		"""Instance of the KeyRingBase."""
		if password is None:
			password = os.environ['CRYPT_KEYRING_TOKEN']
		self.password = password

		self._active_namespace = namespace or 'default'

		self._kek = None
		self.cache_file = self.directory() / 'keyring.crpt'

		self.key_store = dict()
		self._load()

	def _load(self):
		"""Load the keyring data from disk."""
		if self.cache_file.exists():
			with CryptFileBinaryIO(self.cache_file, 'rb', self.main_key, self.main_key) as fh:
				blob = cfjson.load(fh)
				self.key_store = blob['key_store']

	def _dump(self):
		"""Write the keyring data to disk."""
		temp_file = self.cache_file.with_suffix('.temp')
		try:
			with CryptFileTextIO(temp_file, self.main_key, self.main_key, 'wb') as fh:
				cfjson.dump({'key_store': self.key_store}, fh)
			temp_file.replace(self.cache_file)
		finally:
			temp_file.unlink(missing_ok=True)


class KeyRing(KeyRingBase):
	"""KeyRing helps you keep track of your applications RSA keys in a secure way.

	KeyRing uses a key-value store, where the key is a string and the value is an RSA PrivateKey or PrivateKeySet.
	"""

	def __init__(self, namespace=None, password=None):
		super(KeyRingBase, self).__init__(namespace, password)

	@property
	def namespace(self) -> str:
		"""Return the active namespace."""
		return self._active_namespace

	def set_namespace(self, namespace: str):
		"""Set the active namespace so you don't have to do it per call. If the namespace didn't previously exist it will be created.

		:param namespace: The namespace to mak active.
		"""
		self._active_namespace = namespace

	def attach_key(self, key_name: str, key: RSAPrivateKey, namespace: Optional[str] = None) -> RSAPrivateKey:
		"""Attach an existing RSAPrivate key to the KeyRing.

		:param key_name: The name used to access the key.
		:param key: RSAPrivateKey object.
		:param namespace: Namespace in which the key will be saved.
		"""
		namespace = namespace or self.namespace
		if namespace not in self.key_store:
			self.key_store[namespace] = dict()
		if key_name not in self.key_store[namespace]:
			self.key_store[namespace][key_name] = list()

		self.key_store[namespace][key_name].insert(0, key)
		self._dump()
		return key

	def attach_key_file(
			self, key_name: str, pem_file: PathLike, password: Optional[str] = None,
			namespace: Optional[str] = None) -> RSAPrivateKey:
		"""Attach an existing RSAPrivate key (from disk) to the KeyRing.

		:param key_name: The name used to access the key.
		:param pem_file: Path of the existing PEM file on disk.
		:param password: Password for the PEM file if required.
		:param namespace: Namespace in which the key will be saved.
		"""
		return self.attach_key(key_name, pem_file_to_private_key(pem_file, password), namespace)

	def create_key(self, key_name: str, namespace: Optional[str] = None, key_size: int = 4096) -> RSAPrivateKey:
		"""Create a new key and attach it to the KeyRing.

		:param key_name: key_name: The name used to access the key.
		:param namespace: Namespace in which the key will be saved.
		:param key_size: How big the key should be in bits
		"""
		new_key = generate_private_key(key_size)
		return self.attach_key(key_name, new_key, namespace)

	def delete_key(self, key_name: str, namespace: str):
		"""Deletes the given key with out any further prompt or chance of undo.

		:param key_name: key_name: The name of the key to delete.
		:param namespace: Namespace in which the key is saved.

		WARNING: There are no recovery options. The key will be destroyed and all data encrypted with it will be
		permanently encrypted, effectively destroyed.
		"""
		namespace = namespace or self.namespace
		del self.key_store[namespace][key_name]
		self._dump()

	def roll_key(self, key_name: str, namespace: Optional[str] = None) -> RSAPrivateKey:
		"""Replace the existing key with a new one of the same strength. The old key will still be available in the
		key set for decrypting, but all future encrypting should be done with the new key.

		:param key_name: key_name: The name used to access the key.
		:param namespace: Namespace in which the key will be saved.
		"""
		existing_key = self.get_key(key_name, namespace)
		new_key = generate_private_key(existing_key.key_size)
		self.attach_key(key_name, new_key, namespace)
		return new_key

	def get_key(self, key_name: str, namespace: Optional[str] = None) -> RSAPrivateKey:
		"""Get the first key in the key set for the given key_name.

		:param key_name: key_name: The name used to access the key.
		:param namespace: Namespace in which the key will be saved.
		"""
		namespace = namespace or self.namespace
		return self.key_store[namespace][key_name][0]

	def get_keys(self, key_name: str, namespace: Optional[str] = None) -> RSAPrivateKeySet:
		"""Get the key set for the given key_name.

		:param key_name: key_name: The name used to access the key.
		:param namespace: Namespace in which the key will be saved.
		"""
		namespace = namespace or self.namespace
		return self.key_store[namespace][key_name]


def password_generator(length: int) -> str:
	"""Generate a new access token of the given length.

	:param length: The length of the generated access token.

	:meta private:
	"""
	symbols = r"""!#$%&*+-?@_~"""
	alphabet = string.ascii_letters + string.digits + symbols

	def is_symbol(c):
		return c in symbols

	while True:
		password = ''.join(secrets.choice(alphabet) for _ in range(length))
		if (
				any(c.islower() for c in password)
				and any(c.isupper() for c in password)
				and sum(c.isdigit() for c in password) >= 3
				and sum(is_symbol(c) for c in password) >= 3):
			break
	return password


if __name__ == '__main__':
	# os.environ['']

	# KeyRingBase.create()
	# KeyRingBase.change_main_password()
	# ring = KeyRing(namespace='bank')
	# ring.attach_key_file('encryption_key', Path('c:/source/bank/config/active.pem'))
	# print(ring.get_keys('encryption_key'))

	# print(password_gen(50))

	# keyring = KeyRing(namespace='bank')
	#
	# keyring.key_store['bank']['cache'] = keyring.key_store['bank']['encryption_key']
	# # keyring.key_store['bank']['config'] = keyring.key_store['bank']['encryption_key']
	# keyring.key_store['bank']['crypt_file'] = keyring.key_store['bank']['encryption_key']
	# keyring.key_store['bank']['identity'] = keyring.key_store['bank']['encryption_key']
	#
	# keyring.roll_key('cache')
	# keyring.roll_key('crypt_file')
	# keyring.roll_key('identity')

	# keyring.create_key('cache')
	# keyring.create_key('config')
	# keyring.create_key('crypt_file')
	# keyring.create_key('identity')
	pass



import typing

from cryptography import exceptions
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, utils
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.hashes import HashAlgorithm

from .types import RSAKeySet
from .exceptions import InvalidSignature, EncryptionFailed, DecryptionFailed


def encrypt(message: bytes, public_key: RSAKeySet) -> bytes:
	"""Encrypt the bytes with the key."""
	if not isinstance(public_key, (RSAPrivateKey, RSAPublicKey)):
		public_key = public_key[0]
	if isinstance(public_key, RSAPrivateKey):
		public_key = public_key.public_key()

	try:
		ciphertext = public_key.encrypt(
			message,
			padding.OAEP(
				mgf=padding.MGF1(algorithm=hashes.SHA256()),
				algorithm=hashes.SHA256(),
				label=None
			)
		)
		return ciphertext
	except ValueError:
		pass
	raise EncryptionFailed('Encryption failed')


def decrypt(ciphertext: bytes, private_keys: RSAKeySet) -> bytes:
	"""Decrypt the bytes with the key."""
	if isinstance(private_keys, (RSAPrivateKey, RSAPublicKey)):
		private_keys = (private_keys,)
	pad = padding.OAEP(
				mgf=padding.MGF1(algorithm=hashes.SHA256()),
				algorithm=hashes.SHA256(),
				label=None
			)
	for private_key in private_keys:
		try:
			message = private_key.decrypt(ciphertext, pad)
			return message
		except ValueError:
			pass
	raise DecryptionFailed('Decryption failed')


def sign(digest: bytes, private_key: RSAKeySet, chosen_hash: typing.Optional[HashAlgorithm] = None) -> bytes:
	"""Generate a signature for the bytes with the key."""
	if not isinstance(private_key, (RSAPrivateKey, RSAPublicKey)):
		private_key = private_key[0]

	if chosen_hash is None:
		chosen_hash = hashes.SHA256()

	return private_key.sign(
		digest,
		padding.PSS(
			mgf=padding.MGF1(hashes.SHA256()),
			salt_length=padding.PSS.MAX_LENGTH
		),
		utils.Prehashed(chosen_hash)
	)


def verify(
		digest: bytes, sig: bytes, public_keys: RSAKeySet,
		chosen_hash: typing.Optional[HashAlgorithm] = None) -> None:
	"""Verify the given signature matches the bytes."""
	if isinstance(public_keys, (RSAPrivateKey, RSAPublicKey)):
		public_keys = (public_keys,)

	if chosen_hash is None:
		chosen_hash = hashes.SHA256()
	pad = padding.PSS(
					mgf=padding.MGF1(hashes.SHA256()),
					salt_length=padding.PSS.MAX_LENGTH
				)
	pre_hash = utils.Prehashed(chosen_hash)

	for public_key in public_keys:
		try:
			public_key.verify(sig, digest, pad, pre_hash)
			return
		except exceptions.InvalidSignature:
			pass
	raise InvalidSignature('Signature verification failed')



import array
import mmap
import os
from typing import TypeVar

ALGO_AES = 1  #: AES algorithm
ALGO_MODE_CBC = 1  #: Algorithm mode Cipher Block Chaining  # Not implemented
ALGO_MODE_CTR = 2  #: Algorithm mode Counter
PAD_NONE = 0  #: Padding algorithm None
PAD_PKCS7 = 1  #: Padding algorithm PKCS#7  # Not implemented
HASH_SHA256 = 1  #: Hashing algorithm SHA256
MAC_NONE = 0  #: Message authentication code None
MAC_HMAC = 1  #: Message authentication code HMAC
MAC_RSA_SIGNED = 2  #: Message authentication code RSA signature
constant_names = {
	'algorithm': {
		ALGO_AES: 'AES',
	},
	'algorithm_mode': {
		ALGO_MODE_CBC: 'CBC',
		ALGO_MODE_CTR: 'CTR',
	},
	'algorithm_pad': {
		PAD_NONE: 'None',
		PAD_PKCS7: 'PKCS7',
	},
	'algorithm_mac': {
		MAC_NONE: 'None',
		MAC_HMAC: 'HMAC',
		MAC_RSA_SIGNED: 'RSA signed',
	},
	'algorithm_mac_hash': {
		HASH_SHA256: 'SHA256',
	},
}


def parameter_to_name(parameter, parameter_value):
	return constant_names[parameter].get(parameter_value, parameter_value)


PathLike = TypeVar("PathLike", str, bytes, os.PathLike)
ReadableBuffer = TypeVar("ReadableBuffer", bytes, bytearray, memoryview, array.array, mmap.mmap)
WriteableBuffer = TypeVar("WriteableBuffer", bytes, bytearray, memoryview, array.array, mmap.mmap)

SIZE_8MiB = 1024 * 1024 * 8



import os
import struct
from collections import OrderedDict
from typing import BinaryIO

from .constants import MAC_NONE, PathLike
from .exceptions import FileParserError


def is_crypt_file(filename: PathLike) -> bool:
	"""True if the file looks like a crpt file."""
	with open(filename, 'rb') as fh:
		return fh.read(4) == ContainerFormat.magic_bytes


class ContainerFormat(object):
	"""Header for a binary container file format for storing data encrypted."""

	magic_bytes = b'CRYP'  #: Magic byte values

	file_type = (
		('magic_bytes', '4s'),  # 4 Char
		('format_version', 'B'),  # uchar
	)  #: File type fields and data types
	file_type_format = '<' + ''.join(b for _, b in file_type)  #: File type struct definition

	@staticmethod
	def create(*args, **kwargs) -> "ContainerFormat":
		"""Construct a ContainerFormat object for a new file."""
		versions = ContainerFormat.get_supported_versions()
		klass = versions[max(versions.keys())]
		return klass(*args, **kwargs)

	@staticmethod
	def create_from_file(fh: BinaryIO) -> "ContainerFormat":
		"""Create ContainerFormat object from an existing file."""
		fh.seek(0)
		klass = ContainerFormat.get_version_class(fh)

		size = struct.calcsize(klass.format)
		header_bytes = fh.read(size)[:size]

		if len(header_bytes) != size:
			raise RuntimeError('Failed to read the whole header.')

		# Unpack the header
		fields = struct.unpack(klass.format, header_bytes)
		header = klass(*fields, file_handle=fh)
		return header

	@classmethod
	def get_version_class(cls, fh: BinaryIO) -> 'ContainerFormat':
		"""Returns the child class that is registered to handle the version number from the file.

		:raises
			FileParserError: There is an error parsing and encrypted file.

		:meta private:
		"""
		size = struct.calcsize(cls.file_type_format)
		check_bytes = fh.read(size)[:size]
		magic_bytes, format_version = struct.unpack(cls.file_type_format, check_bytes)
		if magic_bytes != cls.magic_bytes:
			raise FileParserError('Bad magic number')

		versions = cls.get_supported_versions()
		if format_version not in versions:
			raise FileParserError(
				'Unknown version number. The file could be corrupt, or created by a newer version of '
				'this software.')

		return versions[format_version]

	@classmethod
	def get_supported_versions(cls):
		"""Queries all the sub-classes for the format versions they represent.

		:meta private:
		"""
		return {subclass.format_version: subclass for subclass in cls.__subclasses__()}

	def __init__(self):
		# The rest are offset/length pairs to allow for variable length.
		self.file_size = None

	def __getattr__(self, item):
		if item in self.fields:
			return self.fields[item]

	def __getitem__(self, item):
		return self.fields[item]

	def __bytes__(self):
		return struct.pack(ContainerFormat.file_type_format, self.magic_bytes, self.format_version)

	def __len__(self):
		return struct.calcsize(ContainerFormat.file_type_format)

	def length(self):
		"""The length in bytes when serialized into a struct."""
		return struct.calcsize(ContainerFormat.file_type_format)

	def to_bytes(self):
		"""Serialize into a struct."""
		return bytes(self)

	def load_variable_size_fields(self, fh):
		"""Parse the file and unpack any know variable length data described in the header.

		:meta private:
		"""
		return NotImplemented

	def set_file_size(self, file_size):
		"""Set the size of the file on disk.

		This is used to calculate some missing lengths and offsets for the data payload and signature.
		"""
		self.file_size = file_size

	@property
	def data_length(self):
		"""The length of the data payload in bytes.

		This is usually only available after the file size is set.

		:meta private:
		"""
		return NotImplemented


class ContainerFormatV1(ContainerFormat):
	labels = (
		('algorithm', 'B'),  # uchar
		('algorithm_mode', 'B'),  # uchar
		('algorithm_pad', 'B'),  # uchar
		('algorithm_mac', 'B'),  # uchar
		('algorithm_mac_hash', 'B'),  # uchar
		('chunk_size', 'i'),  # int

		('nonce_offset', 'I'),  # uint
		('nonce_length', 'I'),  # uint
		('key_offset', 'I'),  # uint
		('key_length', 'I'),  # uint

		('data_offset', 'I'),  # uint

		('mac_length', 'h'),  # short int
	)  #: Fields and data types

	format = '<' + ''.join(b for a, b in labels)  #: Struct definition

	format_version = 1  #: Registered version number

	def __init__(self, *args, file_handle=None, **kwargs):
		super(ContainerFormatV1, self).__init__()
		# Fields are packed into the header
		self.fields = OrderedDict(zip([a for a, b in self.labels], args))

		self.nonce = None
		self.key_material = None
		self.mac = None

		for field, value in kwargs.items():
			self.fields[field] = value

		if file_handle:
			self.load_variable_size_fields(file_handle)

	def __bytes__(self):
		self.update_offsets()
		field_values = [self.fields[label] for label, _ in self.labels]
		return super(ContainerFormatV1, self).__bytes__() + struct.pack(self.format, *field_values)

	def __len__(self):
		return super(ContainerFormatV1, self).__len__() + struct.calcsize(self.format)

	def length(self):
		return len(self)

	def load_variable_size_fields(self, fh):
		# Use the header info to get the nonce
		fh.seek(self.nonce_offset)
		self.set_nonce(fh.read(self.nonce_length))

		# Use the self info to get the encrypted key material
		fh.seek(self.key_offset)
		self.set_key_material(fh.read(self.key_length))

		# Use the self info to get the mac
		if self.mac_length != MAC_NONE:
			fh.seek(self.mac_end_offset, os.SEEK_END)
			self.set_mac(fh.read(self.mac_length))

	def update_offsets(self):
		head = self.length()
		self.fields['nonce_offset'] = head
		head += self.fields['nonce_length']

		self.fields['key_offset'] = head
		head += self.fields['key_length']

		self.fields['data_offset'] = head

	def set_nonce(self, nonce):
		self.nonce = nonce
		self.fields['nonce_length'] = len(nonce)

	def set_key_material(self, key_material):
		self.key_material = key_material
		self.fields['key_length'] = len(key_material)

	def set_mac(self, code):
		self.mac = code
		self.fields['mac_length'] = len(code)

	def set_mac_length(self, mac_length):
		self.fields['mac_length'] = mac_length

	@property
	def mac_offset(self):
		return self.file_size + self.mac_end_offset

	@property
	def mac_end_offset(self):
		# Relative to the end of file so we don't need to know the data length to recover it.
		return -self.mac_length

	@property
	def data_length(self):
		# Calculate this dynamically so we don't need to know it before we start writing the header.
		return self.file_size - self.mac_length - self.data_offset



from cfcrypt.crypt_file import CryptFileBinaryIO, CryptFileTextIO
from cfcrypt.file_format import is_crypt_file

import cfcrypt.file_utils

from cfcrypt.exceptions import CryptException, FileParserError
from cfcrypt.rsa.exceptions import RSAException, EncryptionFailed, DecryptionFailed, InvalidSignature


def register_serde():
	"""Registers with kson serialization."""
	def rsa_private_key_decode(dct):
		"""Decode json dict into class objects."""
		from cfcrypt.rsa.key_helpers import pem_to_private_key
		cls_name = dct['__json_type__']
		if cls_name in ('_RSAPrivateKey',):
			return pem_to_private_key(dct['private_key'].encode())
		raise TypeError()

	def rsa_private_key_encode(obj):
		"""Encode the object into a json safe dict."""
		from cfcrypt.rsa.key_helpers import private_key_to_pem
		return {
			'__json_type__': type(obj).__name__,
			'private_key': private_key_to_pem(obj).decode('utf-8')
		}

	JsonTypeRegister.register('_RSAPrivateKey', rsa_private_key_encode, rsa_private_key_decode)


try:
	from kson.kson import JsonTypeRegister
	register_serde()
except ImportError:
	pass


__all__ = (
	'is_crypt_file', 'CryptFileBinaryIO', 'CryptFileTextIO',
	'CryptException', 'RSAException', 'EncryptionFailed', 'DecryptionFailed', 'InvalidSignature')



from typing import TypeVar, Iterable, Union

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

RSAKey = TypeVar('RSAKey', RSAPrivateKey, RSAPublicKey)
RSAPrivateKeySet = TypeVar("RSAPrivateKeySet", Iterable[RSAPrivateKey], RSAPrivateKey)
RSAPublicKeySet = TypeVar("RSAPublicKeySet", Iterable[RSAPublicKey], RSAPublicKey)
RSAKeySet = TypeVar("RSAKeySet", Iterable[Union[RSAPrivateKey, RSAPublicKey]], Union[RSAPrivateKey, RSAPublicKey])



"""Helpers for working with RSA keys."""
import os
import typing
from pathlib import Path

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

from .types import RSAPrivateKeySet, RSAPublicKeySet

PathLike = typing.TypeVar("PathLike", str, bytes, os.PathLike)


#
# Create
#

def generate_private_key(key_size: int = 2048) -> RSAPrivateKey:
	"""Create a new RSAPrivateKey"""
	return rsa.generate_private_key(
			public_exponent=65537,
			key_size=key_size,
	)


def generate_private_pem(key_size: int = 2048, password: typing.Optional[str] = None) -> bytes:
	"""Create a new RSAPrivateKey encoded into PEM bytes"""
	private_key = generate_private_key(key_size)
	pem = private_key_to_pem(private_key, password)
	return pem


def generate_keypair(key_size: int = 2048) -> typing.Tuple[RSAPrivateKey, RSAPublicKey]:
	"""Create a new RSAPrivateKey and its matching RSAPublicKey"""
	private_key = generate_private_key(key_size)
	public_key = private_key.public_key()
	return private_key, public_key


#
# Transform
#

def pem_to_private_key(key_bytes, password: typing.Optional[str] = None) -> RSAPrivateKey:
	"""Load PEM bytes into an RSAPrivateKey"""
	if password is not None:
		password = password.encode()
	private_key = serialization.load_pem_private_key(key_bytes, password=password)
	return private_key


def pem_to_keypair(
		key_bytes: bytes, password: typing.Optional[str] = None) -> typing.Tuple[RSAPrivateKey, RSAPublicKey]:
	"""Load PEM bytes into an RSAPrivateKey and its matching RSAPublicKey"""
	private_key = pem_to_private_key(key_bytes, password)
	return private_key, private_key.public_key()


def pem_file_to_private_key(filepath: PathLike, password: typing.Optional[str] = None):
	"""Load PEM bytes from a file into an RSAPrivateKey"""
	filepath = Path(filepath)
	with filepath.open('rb') as fh:
		return pem_to_private_key(fh.read(), password)


def pem_file_to_keypair(
		filepath: PathLike, password: typing.Optional[str] = None) -> typing.Tuple[RSAPrivateKey, RSAPublicKey]:
	"""Load PEM bytes from a file into an RSAPrivateKey and its matching RSAPublicKey"""
	private_key = pem_file_to_private_key(filepath, password)
	return private_key, private_key.public_key()


def private_key_to_pem(private_key: RSAPrivateKey, password: typing.Optional[str] = None) -> bytes:
	"""Serialize an RSAPrivateKey into PEM bytes"""
	if password:
		password = password.encode()
		encryption_algorithm = serialization.BestAvailableEncryption(password)
	else:
		encryption_algorithm = serialization.NoEncryption()

	return private_key.private_bytes(
	   encoding=serialization.Encoding.PEM,
	   format=serialization.PrivateFormat.PKCS8,
	   encryption_algorithm=encryption_algorithm
	)


def private_key_to_pem_file(
		filepath: PathLike, private_key: RSAPrivateKey, password: typing.Optional[str] = None) -> Path:
	"""Serialize an RSAPrivateKey into PEM bytes and write them into a file."""
	filepath = Path(filepath)
	with filepath.open('wb') as fh:
		fh.write(private_key_to_pem(private_key, password))
	return filepath


#
# Pack into KeySets
#

def to_public_key(key_set: typing.Union[RSAPublicKeySet, RSAPrivateKeySet]) -> RSAPublicKey:
	"""Extract a single public key from the key set."""
	key = key_set
	if not isinstance(key_set, (RSAPrivateKey, RSAPublicKey)):
		key = key_set[0]
	if isinstance(key, RSAPrivateKey):
		return key.public_key()
	return key


def to_public_key_set(key_set: RSAPrivateKeySet) -> RSAPublicKeySet:
	"""Recover the RSAPublic key from each RSAPrivateKey in the set."""
	if isinstance(key_set, (RSAPrivateKey, RSAPublicKey)):
		key_set = [key_set, ]
	return [key.public_key() for key in key_set]



"""Base class for EncryptedIO implementations."""

import io
import warnings
from pathlib import Path
from typing import BinaryIO, Iterable

from .constants import parameter_to_name, PathLike, ReadableBuffer
from .file_format import ContainerFormat

from .rsa import encrypt, decrypt
from .rsa.key_helpers import to_public_key
from .rsa.types import RSAPrivateKeySet


class EncryptedIOBase(io.RawIOBase):
	"""Provides a frame work for reading and writing to and encrypted container file.

	Sub classes should implement the stubs to provide the actual encryption.
	"""

	def __init__(self, filename: PathLike, mode: str, encryption_key: RSAPrivateKeySet, verify: bool = True):
		self.suppress_verification_warning = True

		self.encryption_key_private = encryption_key

		self.read_only = mode == 'rb'
		self.write_only = mode == 'wb'

		self.filename = filename
		if self.read_only:
			self.fh = open(filename, 'rb')
		elif self.write_only:
			self.fh = open(filename, 'wb')
		else:
			raise RuntimeError('File mode not supported.')

		self.block_size = 16
		self.header = None

		self.key = None
		self.nonce = None
		self.counter = 0

		self.data_offset = None
		self.data_length = None

		self.head = 0
		self.block_offset = 0

		self.verification_complete = False
		if self.readable():
			self._prepare_for_reading(verify)
		elif self.writable():
			self._prepare_for_writing()
		else:
			raise RuntimeError('Only reading or writing supported.')

		self.suppress_verification_warning = False

	@property
	def _encryption_config(self) -> dict:
		"""The settings used to setup the encryption library."""
		return dict()

	def _check_encryption_config(self, parameter: str):
		"""Check that the encryption settings the file was generated with match the active encryption settings."""
		hp = self.header[parameter]
		cp = self._encryption_config[parameter]
		if hp != cp:
			raise RuntimeError(
				f'Mismatched parameter: {parameter}. The active encryption setting is {parameter_to_name(parameter, cp)} but '
				f'the file was encrypted with {parameter_to_name(parameter, hp)}.')

	def _prepare_for_reading(self, verify: bool):
		"""Read the header from the file. Get the state needed to decrypt. Run the verification check for file tampering."""
		self.header = self._read_file_header(self.fh)

		for parameter in ('algorithm', 'algorithm_mode', 'algorithm_pad', 'algorithm_mac', 'algorithm_mac_hash'):
			self._check_encryption_config(parameter)

		file_size = Path(self.filename).stat().st_size
		self.header.set_file_size(file_size)

		self.key = self._decrypt_symmetric_key()
		self.nonce = self.header.nonce
		self.data_offset = self.header.data_offset
		self.data_length = self.header.data_length

		if verify:
			self.verify()

		self.fh.seek(self.data_offset)

	def _prepare_for_writing(self):
		"""Setup our state needed for encrypting data. Create the file header and write it to the stream."""
		self.key, self.nonce = self.generate_key_material()
		encrypted_key_material = self._encrypted_symmetric_key()

		self.header = self._create_header()
		self.header.set_nonce(self.nonce)
		self.header.set_key_material(encrypted_key_material)
		self.header.set_mac_length(self.mac.length)

		header_bytes = self.header.to_bytes()
		self._write(header_bytes)
		self._write(self.nonce)
		self._write(encrypted_key_material)

	def _write(self, data: bytes):
		"""Tee the write operation to all the places it needs to go."""
		self.fh.write(data)
		if self.mac:
			self.mac.update(data)

	#
	# Interface
	#

	# Support context manager and therefore supports the with statement. In this example, file is closed after the with
	# statement’s suite is finished—even if an exception occurs
	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, exc_traceback):
		self.suppress_verification_warning = exc_type is not None
		self.close()

	def __iter__(self) -> bytes:
		if not self.readable():
			raise io.UnsupportedOperation('This file is not open read mode.')

		read_length = self.block_offset + self.block_size
		buffer = self.read(read_length)
		while len(buffer) > 0:
			for char in buffer:
				yield char.to_bytes(1, 'big')
			buffer = self.read(self.block_size)

	def close(self):
		"""Flush and close this stream. This method has no effect if the stream is already closed. If the stream has not
		already been finalized, it will now be finalized."""
		if self.fh.closed:
			return

		if self.readable():
			if not self.verification_complete and not self.suppress_verification_warning:
				warnings.warn('This file has not been verified. Do not trust the contents until you run a successful verification.')

		if self.writable():
			if not self.finalized:
				self.finalize()

		self.fh.close()

	@property
	def closed(self) -> bool:
		"""True if the stream is closed."""
		return self.fh.closed

	# noinspection SpellCheckingInspection
	def fileno(self) -> int:
		"""Return the underlying file descriptor (an integer) of the stream if it exists. An OSError is raised if the
		IO object does not use a file descriptor. """
		return self.fh.fileno()

	def finalize(self):
		"""Complete the writing operations by appending a cryptographic signature. Once finalized you can no longer
		write more data to the stream."""
		return NotImplemented

	@property
	def finalized(self) -> bool:
		"""True if the stream has been finalized."""
		return False

	def flush(self):
		"""Flush the write buffers of the stream if applicable. This does nothing for read-only and non-blocking streams."""
		return self.fh.flush()

	# noinspection SpellCheckingInspection
	def isatty(self) -> bool:
		"""Return True if the stream is interactive (i.e., connected to a terminal/tty device)."""
		return False

	def read(self, size=-1) -> bytes:
		"""Read up to size bytes from the object and return them. As a convenience, if size is unspecified or -1,
		all bytes until EOF are returned."""
		if not self.readable():
			raise io.UnsupportedOperation('This file is not open read mode.')

		if size < 0 or size is None:
			size = max(self.data_length - self.head, 0)
		else:
			size = size

		# We need to read from the start of a block to make the decryption work.
		read_size = self.block_offset + size

		decryptor = self.decryptor
		encrypted_data = self.fh.read(read_size)
		data = decryptor.update(encrypted_data) + decryptor.finalize()

		# Don't give the user the extra data between the block start and the head position
		data = data[self.block_offset:]

		# Don't give them more bytes than they asked for.
		data = data[:size]

		# Trim off any data after the edge of the container boundary.
		read_size = len(data)
		if self.head + read_size > self.data_length:
			overshoot = self.head + read_size - self.data_length
			data = data[:read_size - overshoot]
		self.head += len(data)

		# The seek here is important. It puts the file head on a block boundary and sets up
		# the self.block_offset value for the next read.
		self.seek(self.head)
		return data

	def readable(self) -> bool:
		"""Return True if the stream can be read from."""
		return self.read_only

	# noinspection SpellCheckingInspection
	def readline(self, size=-1):
		r"""Read and return one line from the stream. If size is specified, at most size bytes will be read.

		The line terminator is always b'\\n' for binary files; for text files, the newline argument to open() can be
		used to select the line terminator(s) recognized.
		"""
		return NotImplemented

	# noinspection SpellCheckingInspection
	def readlines(self, hint=-1):
		"""
		Read and return a list of lines from the stream. hint can be specified to control the number of lines read:
		no more lines will be read if the total size (in bytes/characters) of all lines so far exceeds hint.

		Note that it’s already possible to iterate on file objects using for line in file: ... without calling
		file.readlines().
		"""
		return NotImplemented

	def seek(self, offset, whence=io.SEEK_SET) -> int:
		"""Change the stream position to the given byte offset. offset is interpreted relative to the position indicated
		by whence. The default value for whence is SEEK_SET. Values for whence are:

			SEEK_SET or 0 – start of the stream (the default); offset should be zero or positive
			SEEK_CUR or 1 – current stream position; offset may be negative
			SEEK_END or 2 – end of the stream; offset is usually negative

		Return the new absolute position."""
		if not self.seekable():
			raise io.UnsupportedOperation('seek')

		if whence == io.SEEK_SET:
			# Relative to the start
			self.head = offset
		elif whence == io.SEEK_CUR:
			# Relative to the current position
			self.head = self.head + offset
		elif whence == io.SEEK_END:
			# Relative to the end
			self.head = self.data_length - 1 + offset
		else:
			raise NotImplementedError(f"Whence of '{whence}' is not supported.")

		self.head = self._clamp_to_file_bounds(self.head)
		self.counter = self.head // self.block_size
		self.block_offset = self.head % self.block_size
		block_start = self.counter * self.block_size

		file_offset = self.data_offset + block_start
		if self.fh.tell() != file_offset:
			self.fh.seek(self.data_offset + block_start)
		return self.head

	def seekable(self) -> True:
		"""Return True if the stream supports random access."""
		return self.read_only

	def tell(self) -> int:
		"""Return the current stream position."""
		return self.head

	def truncate(self, size=None):
		"""Not supported."""
		raise io.UnsupportedOperation('truncate')

	def writable(self) -> bool:
		"""Return True if the stream supports writing."""
		return self.write_only

	def write(self, data: ReadableBuffer):
		"""Write the given data to the stream."""
		if not self.writable():
			raise io.UnsupportedOperation('This file is not open write mode.')
		if self.finalized:
			raise io.UnsupportedOperation('This file has already been finalized.')

		encrypted_data = self.encryptor.update(data)
		self._write(encrypted_data)

	def writelines(self, lines: Iterable[ReadableBuffer]):
		"""Write a list of lines to the stream. Line separators are not added, so it is usual for each of the lines
		provided to have a line separator at the end."""
		return NotImplemented

	def verify(self):
		"""Check the file signature matches the given identity key. Verify that file has not been modified since creation."""
		return NotImplemented

	#
	# Internal
	#

	def _clamp_to_file_bounds(self, position: int) -> int:
		"""Make sure our read and seek don't go outside the bounds of the data payload."""
		return min(max(0, position), self.data_length)

	def _create_header(self) -> ContainerFormat:
		"""Return a header object."""
		return NotImplemented

	def generate_key_material(self):
		"""Generate the required symmetric key material.
		The values should come from a strong random source like os.urandom()

		WARNING: This value MUST NOT be hardcoded.
		"""
		return NotImplemented

	def _read_file_header(self, fh: BinaryIO):
		"""Read the file header and return an object representation."""
		return NotImplemented

	@property
	def decryptor(self):
		"""Return a fresh decryptor."""
		return NotImplemented

	@property
	def encryptor(self):
		return NotImplemented

	@property
	def mac(self):
		return NotImplemented

	@property
	def signature_hash(self):
		return NotImplemented

	def _encrypted_symmetric_key(self) -> bytes:
		"""Return the symmetric key after it has been encrypted with the rsa_key.

		The encrypted_key will be written into the file. The private_key will need to be used to recover the contents.
		"""
		try:
			return encrypt(self.key, to_public_key(self.encryption_key_private))
		except ValueError:
			raise RuntimeError('Failed to encrypt the symmetric key. The RSA public key may not be long enough.')

	def _decrypt_symmetric_key(self) -> bytes:
		"""Return the symmetric key after it has been decrypted with the private_key.

		The symmetric will be used to decrypt the file contents.
		"""
		encrypted_key_material = self.header.key_material
		return decrypt(encrypted_key_material, self.encryption_key_private)



"""Provides a container format and implementation to store files on disk encrypted."""

import io
import os

import typing
from cryptography.hazmat.primitives import ciphers
from cryptography.hazmat.primitives.ciphers import algorithms, modes

from ._crypt_file import EncryptedIOBase
from .file_format import ContainerFormat
from .message_authentication import MessageAuthenticatorRSA

from .constants import ALGO_AES, ALGO_MODE_CTR, PAD_NONE, HASH_SHA256, MAC_RSA_SIGNED, SIZE_8MiB, PathLike
from .rsa.types import RSAPrivateKeySet

__all__ = ['CryptFileTextIO', 'CryptFileBinaryIO']


class CryptFileBinaryIO(EncryptedIOBase):
	"""Implementation of EncryptedIO using AES-CTR

	Writing:
		The rsa_key is used to lock the file after writing, providing secrecy of the symmetric keys.
		The private_key is used to sign the file after writing, providing material for Integrity and Authentication.

	Reading:
		The private_key is used to unlock the file for reading, removing the secrecy of the symmetric keys.
		The rsa_key is used to verify the signature, proving Integrity and Authentication.
	"""

	def __init__(self, filename: PathLike, mode, encryption_key: RSAPrivateKeySet, signing_key: RSAPrivateKeySet, *args, **kwargs):

		self._config = {
			'algorithm': ALGO_AES,
			'algorithm_mode': ALGO_MODE_CTR,
			'algorithm_pad': PAD_NONE,
			'algorithm_mac': MAC_RSA_SIGNED,
			'algorithm_mac_hash': HASH_SHA256,
		}

		self.signing_key_private = signing_key

		self.encryptor_finalized = False

		self._encryptor = None
		self.mac_object = None

		super(CryptFileBinaryIO, self).__init__(filename, mode, encryption_key, *args, **kwargs)

	@property
	def _encryption_config(self):
		return self._config

	@property
	def finalized(self):
		return self.writable() and self.encryptor_finalized

	@property
	def cipher(self):
		"""Construct and return the Cipher object base on the encryption config.

		Note: We should **not** cache the cipher object. We need to issue a fresh instance every time self.counter changes.

		:meta private:
		"""
		if self.header.algorithm != self._encryption_config['algorithm']:
			raise RuntimeError('Encryption algorithm not supported')
		if self.header.algorithm_mode != self._encryption_config['algorithm_mode']:
			raise RuntimeError('Encryption mode not supported')

		# Note: We need to keep the nonce in sync with the file position. This gets out of sync if we seek around the
		# file, so we have to manually re-sync it.
		nonce = add_int_to_bytes(self.nonce, self.counter, self.block_size)
		algorithm = algorithms.AES(self.key)
		algorithm_mode = modes.CTR(nonce)
		return ciphers.Cipher(algorithm, algorithm_mode)

	@property
	def mac(self):
		"""Construct and return the message authenticator.

		:meta private:
		"""
		if self.header.algorithm_mac != self._encryption_config['algorithm_mac']:
			raise RuntimeError('MAC algorithm not supported')
		if self.header.algorithm_mac_hash != self._encryption_config['algorithm_mac_hash']:
			raise RuntimeError('Hash algorithm not supported')
		if not self.mac_object:
			self.mac_object = MessageAuthenticatorRSA(self.signing_key_private)
		return self.mac_object

	@property
	def decryptor(self):
		"""Return a fresh decryptor.

		NOTE: It's important that we use a fresh decryptor after any seek operation. This gives us a chance to update
		the counter value to match the seek location. For contagious reads this is not an issue as it gets managed
		internally to the cipher.

		:meta private:
		"""
		return self.cipher.decryptor()

	@property
	def encryptor(self):
		"""Return the active encryptor.

		Note: It's important that we use the same encryptor when writing to the file. We need the internal nonce values
		to stay in sync so we can decrypt the stream later.

		:meta private:
		"""
		if self.finalized:
			raise RuntimeError('This file has been finalized.')
		if self._encryptor is None:
			self._encryptor = self.cipher.encryptor()
		return self._encryptor

	def _read_file_header(self, fh: io.FileIO):
		"""Construct a file header object from the given file handle."""
		return ContainerFormat.create_from_file(fh)

	def verify(self):
		"""Check the file signature matches the given identity key. Verify that file has not been modified since creation."""
		mac = self.mac
		head = self.tell()
		try:
			self.fh.seek(0)

			mac_offset = self.header.mac_offset
			mac_length = self.header.mac_length

			data_end = mac_offset

			bytes_read = 0
			while True:
				chunk = max(min(data_end - bytes_read, SIZE_8MiB), 0)
				data = self.fh.read(chunk)
				if not data:
					break
				bytes_read += len(data)

				mac.update(data)

			self.fh.seek(mac_offset)
			signature = self.fh.read(mac_length)
			self.mac.verify(signature)

		finally:
			self.seek(head)
			self.verification_complete = True

		return True

	def _create_header(self):
		"""Construct a file header object for a new file."""

		chunk_size = -1

		header = ContainerFormat.create(
			self._encryption_config['algorithm'], self._encryption_config['algorithm_mode'], self._encryption_config['algorithm_pad'],
			self._encryption_config['algorithm_mac'], self._encryption_config['algorithm_mac_hash'], chunk_size,
			nonce_length=len(self.nonce),
			key_length=len(self._encrypted_symmetric_key()),
			)

		return header

	def generate_key_material(self):
		"""Generate the symmetric key material for the file encryption.

		:meta private:
		"""
		key_material = os.urandom(32)
		nonce = os.urandom(16)
		return key_material, nonce

	def finalize(self):
		"""Complete the writing operations by appending a cryptographic signature. Once finalized you can no longer
		write more data to the stream."""
		if not self.writable():
			raise io.UnsupportedOperation('This file is not open write mode.')
		if self.finalized:
			raise io.UnsupportedOperation('This file has already been finalized.')

		encrypted_data = self.encryptor.finalize()
		self.fh.write(encrypted_data)

		if self.mac:
			authentication = self.mac.finalize()
			self.fh.write(authentication)

		self.fh.flush()
		self.fh.close()

		self.encryptor_finalized = True


class CryptFileTextIO(io.TextIOWrapper):
	"""A buffered text stream wrapper on CryptFileBinaryIO."""

	def __init__(self, filename: PathLike, encryption_key: RSAPrivateKeySet, signing_key: RSAPrivateKeySet,
				 mode: str = 'r', verify=True, encoding=None, errors=None, newline=None, line_buffering=False,
				 write_through=False):

		if encryption_key is None:
			raise TypeError('CryptFileTextIO expected an encryption_key')
		if encryption_key is None:
			raise TypeError('CryptFileTextIO expected a signing_key')

		if mode in ('r', 'rb'):
			mode = 'rb'
		elif mode in ('w', 'wb'):
			mode = 'wb'
		else:
			raise io.UnsupportedOperation(f'Mode "{mode}" not supported. Mode must be one of "r", "w"')

		buffer = CryptFileBinaryIO(filename, mode, encryption_key, signing_key, verify)
		buffer = typing.cast(typing.IO[bytes], buffer)
		super(CryptFileTextIO, self).__init__(buffer, encoding, errors, newline, line_buffering, write_through)


def add_int_to_bytes(b, i, block_size):
	"""Add an integer to a byte string."""
	MAX = int.from_bytes(b'\xff' * block_size, byteorder='big') + 1
	# If the counter overflows, it wraps back to zero
	i = (int.from_bytes(b, byteorder='big') + i) % MAX
	return i.to_bytes(block_size, 'big')



import shutil
from pathlib import Path
from typing import Iterable

from . import is_crypt_file, CryptFileBinaryIO
from .constants import PathLike, SIZE_8MiB
from cfcrypt.rsa.types import RSAPrivateKeySet


def encrypt_folder(
		folder_path: PathLike, encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet,
		buffer_size: int = SIZE_8MiB, recursive: bool = False):
	"""
	Encrypt all files in the given folder.

	:param folder_path: Folder to search for files in.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	:param recursive: If True check sub-folders for files as well.
	"""
	folder_path = Path(folder_path)

	if recursive:
		filepaths = folder_path.glob('**/*')
	else:
		filepaths = folder_path.glob('*')

	filepaths = [f for f in filepaths if f.is_file()]
	encrypt_files(filepaths, encryption_key, identity_key, buffer_size)


def encrypt_files(
		filepaths: Iterable[PathLike], encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet,
		buffer_size: int = SIZE_8MiB):
	"""
	Encrypt a list of files in place.

	:param filepaths: Files to encrypt.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	"""
	for filepath in filepaths:
		if is_crypt_file(filepath):
			continue
		encrypt_file(filepath, encryption_key, identity_key, buffer_size)


def encrypt_file(
		filepath: PathLike, encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet, buffer_size: int = SIZE_8MiB):
	"""
	Encrypt a single file in place.

	:param filepath: File to encrypt.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	"""
	filepath = Path(filepath)
	if is_crypt_file(filepath):
		raise RuntimeError('File is already encrypted.')

	output_filepath = filepath.with_suffix('.encrypting')
	try:
		with filepath.open('rb') as input_fh, \
				CryptFileBinaryIO(output_filepath, 'wb', encryption_key, identity_key) as output_fh:
			while True:
				chunk = input_fh.read(buffer_size)
				if len(chunk) == 0:
					break
				output_fh.write(chunk)
		output_filepath.replace(filepath)

	finally:
		output_filepath.unlink(missing_ok=True)


def decrypt_folder(
		folder_path: PathLike, encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet,
		buffer_size: int = SIZE_8MiB, recursive: bool = False):
	"""
	Decrypt all files in the given folder.

	:param folder_path: Folder to search for files in.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	:param recursive: If True check sub-folders for files as well.
	"""
	folder_path = Path(folder_path)

	if recursive:
		filepaths = folder_path.glob('**/*')
	else:
		filepaths = folder_path.glob('*')

	filepaths = [f for f in filepaths if f.is_file()]
	decrypt_files(filepaths, encryption_key, identity_key, buffer_size)


def decrypt_files(
		filepaths: Iterable[PathLike], encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet,
		buffer_size: int = SIZE_8MiB):
	"""
	Decrypt a list of files in place.

	:param folder_path: Files to encrypt.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	"""
	for filepath in filepaths:
		if not is_crypt_file(filepath):
			continue
		decrypt_file(filepath, encryption_key, identity_key, buffer_size)


def decrypt_file(
		filepath: PathLike, encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet, buffer_size: int = SIZE_8MiB):
	"""
	Decrypt a single file in place.

	:param filepath: File to encrypt.
	:param encryption_key: The encryption key to use.
	:param identity_key: The identity key to use.
	:param buffer_size: Control how much of the file is read into memory each loop.
	"""
	filepath = Path(filepath)
	if not is_crypt_file(filepath):
		raise RuntimeError('File is not encrypted.')

	output_filepath = filepath.with_suffix('.decrypting')
	try:
		with CryptFileBinaryIO(filepath, 'rb', encryption_key, identity_key) as input_fh, \
			output_filepath.open('wb') as output_fh:
			while True:
				chunk = input_fh.read(buffer_size)
				if len(chunk) == 0:
					break
				output_fh.write(chunk)
		output_filepath.replace(filepath)

	finally:
		output_filepath.unlink(missing_ok=True)


def secure_file_move(
		source_file: PathLike, target_file: PathLike, encryption_key: RSAPrivateKeySet, identity_key: RSAPrivateKeySet,
		buffer_size: int = SIZE_8MiB):
	"""Move the source to the target and encrypt it."""
	source_file = Path(source_file)
	if is_crypt_file(source_file):
		shutil.move(source_file, target_file)
		return

	try:
		with source_file.open('rb') as input_fh, \
				CryptFileBinaryIO(target_file, 'wb', encryption_key, identity_key) as output_fh:

			while True:
				chunk = input_fh.read(buffer_size)
				if len(chunk) == 0:
					break
				output_fh.write(chunk)

	except Exception:
		target_file.unlink(missing_ok=True)
		raise


def encrypt_data():
	from crypt.key_ring import KeyRing
	ring = KeyRing('bank')
	encryption_key = ring.get_keys('crypt_file')
	identity_key = ring.get_keys('identity')
	encrypt_folder(Path('c:/source/bank/data'), encryption_key, identity_key, True)


def merge():
	from crypt.key_ring import KeyRing
	ring = KeyRing('bank')
	encryption_key = ring.get_keys('crypt_file')
	identity_key = ring.get_keys('identity')
	# decrypt_file(Path('c:/source/bank/scratch/merge.csv'), encryption_key, identity_key)
	# encrypt_file(Path('c:/source/bank/scratch/merge.csv'), encryption_key, identity_key)


if __name__ == '__main__':
	merge()



"""Helpers for working with the cryptography RSA library."""

from .low_level import encrypt, decrypt, sign, verify
from .high_level import encrypt_string, decrypt_string, encrypt_object, decrypt_object
from .types import RSAPrivateKey, RSAPublicKey, RSAKey, RSAPrivateKeySet, RSAPublicKeySet, RSAKeySet
from .exceptions import RSAException, EncryptionFailed, DecryptionFailed, InvalidSignature
from .key_helpers import generate_private_key, generate_private_pem, generate_keypair, pem_to_private_key, pem_to_keypair, pem_file_to_private_key, pem_file_to_keypair, private_key_to_pem, private_key_to_pem_file, to_public_key, to_public_key_set

low_level = ['encrypt', 'decrypt', 'sign', 'verify']
high_level = ['encrypt_string', 'decrypt_string', 'encrypt_object', 'decrypt_object']
# key_helpers = ['generate_private_key', 'generate_private_pem', 'generate_keypair', 'pem_to_private_key', 'pem_to_keypair', 'pem_file_to_private_key', 'pem_file_to_keypair', 'private_key_to_pem', 'private_key_to_pem_file', 'to_public_key', 'to_public_key_set']
# types = ['RSAPrivateKey', 'RSAPublicKey', 'RSAKey', 'RSAPrivateKeySet', 'RSAPublicKeySet']
exceptions = ['RSAException', 'EncryptionFailed', 'DecryptionFailed', 'InvalidSignature']

__all__ = low_level + high_level + exceptions



from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

from .constants import HASH_SHA256

from .rsa.types import RSAPrivateKeySet
from .rsa.low_level import sign, verify
from .rsa.key_helpers import to_public_key_set


class MessageAuthenticator(object):
	"""Unified interface for message authentication codes."""

	def __init__(self):
		self.hash = None
		self._finalized = False
		self._verified = False

	@property
	def finalized(self) -> bool:
		return self._finalized

	@property
	def verified(self) -> bool:
		return self._verified

	@property
	def length(self):
		"""Return the number of bytes the final mac will be."""
		return NotImplemented

	def update(self, data: bytes) -> None:
		"""Update the hash with new data."""
		self.hash.update(data)

	def finalize(self) -> bytes:
		"""Complete the hash and sign it."""
		return NotImplemented

	def verify(self, signature: bytes) -> bytes:
		return NotImplemented


class MessageAuthenticatorRSA(MessageAuthenticator):
	"""RSA authentication implementation."""

	def __init__(self, private_key: RSAPrivateKeySet, hash_type=HASH_SHA256):
		super(MessageAuthenticatorRSA, self).__init__()
		self.private_key = private_key

		if hash_type == HASH_SHA256:
			self.hash = hashes.Hash(hashes.SHA256())
		else:
			raise RuntimeError('Hash type not supported.')

	@property
	def length(self):
		if not isinstance(self.private_key, (RSAPrivateKey, RSAPublicKey)):
			return int(self.private_key[0].key_size / 8)
		return int(self.private_key.key_size / 8)

	def finalize(self):
		if self._finalized:
			return
		self._finalized = True
		signature_hash = self.hash.finalize()
		return sign(signature_hash, self.private_key)

	def verify(self, signature):
		if self._verified:
			return

		signature_hash = self.hash.finalize()
		verify(signature_hash, signature, to_public_key_set(self.private_key))
		self._verified = True
		return self._verified


class MessageAuthenticatorHMAC(MessageAuthenticator):
	"""HMAC-SHA256 implementation."""

	def __init__(self, private_key, hash_type=HASH_SHA256):
		super(MessageAuthenticatorHMAC, self).__init__()

		if not isinstance(private_key, (tuple, list)):
			private_key = [private_key, ]
		self.private_key = private_key

		if hash_type == HASH_SHA256:
			self.hash = hmac.HMAC(private_key[0], hashes.SHA256())
		else:
			raise RuntimeError('Hash type not supported.')

	@property
	def length(self):
		return 32

	def finalize(self):
		if self._finalized:
			return
		self._finalized = True
		return self.hash.finalize()

	def verify(self, signature):
		for key in self.private_key:
			try:
				hmac_hash = hmac.HMAC(key, hashes.SHA256())
				hmac_hash.verify(signature)
				self._verified = True
				return self._verified
			except Exception:
				pass
		raise RuntimeError('Verification failed')
